from rest_framework import exceptions


class RequestException(exceptions.APIException):
    def __init__(self, detail=None, code=None, status_code=None):
        super().__init__(detail, code)
        if status_code:
            self.status_code = status_code
